import abc
from collections import namedtuple


MigrationResult = namedtuple(
    'MigrationResult',
    ('ts', 'num', 'name', 'short', 'long', 'source', 'md5', 'result'))


class DBInterface(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def __init__(self, config):
        pass

    @abc.abstractproperty
    def results(self):
        """ List of all results (MigrationResult)
        """

    @abc.abstractmethod
    def add(self, migration, result):
        """ Add migration result to database
        """

    @abc.abstractmethod
    def find(self, migration):
        """ Fild all results for migration

        Return list of results (MigrationResult).
        """

    @abc.abstractmethod
    def success(self, migration):
        """ Return True if last migration run is success
        """


class DBSimpleBase(DBInterface):
    def __init__(self, config):
        self._config = config

    def find(self, migration):
        return [r for r in self.results
                if r.num == migration.num and r.name == migration.name]

    def success(self, migration):
        results = self.find(migration)

        if not results:
            return False
        else:
            return results[-1].result



import argparse
import configparser
import os
import sys

from migranite import __version__
import migranite.run
import migranite.utils


def _require_config(func):
    def wrapper(parser, args):
        if 'migrations' in args.config:
            func(parser, args)
        else:
            print("Config file {!r} not found".format(args.config['file']),
                  file=sys.stderr)
            sys.exit(1)

    return wrapper


def run_init(parser, args):
    config_path = args.config['file']

    if os.path.exists(config_path):
        print("Config file {!r} is already exists".format(config_path),
              file=sys.stderr)
        sys.exit(1)

    migranite.run.init(args.config, args.migrations, args.templates)


@_require_config
def run_list(parser, args):
    migranite.run.print_list(args.config, args.long, args.all)


@_require_config
def run_create(parser, args):
    if not args.template:
        if 'default' not in args.config['templates']:
            print("Default template not set.", file=sys.stderr)
            sys.exit(1)

        template = args.config['templates']['default']

    template = os.path.join(args.config['templates']['path'], template)

    if not os.path.isfile(template):
        print("{!r} is not exist".format(template), file=sys.stderr)
        sys.exit(1)

    migranite.run.create(args.config, template, args.name[0])


@_require_config
def run(parser, args):
    if args.migrations:
        migranite.run.migrate(
            args.config, args.migrations, args.force)
    else:
        migranite.run.migrate_all(args.config)


def run_help(parser, args):
    parser.print_help()


def run_version(parser, args):
    print(__version__)


def _parse_config(path):
    path = migranite.utils.parse_path(path)

    try:
        with open(path, 'r', encoding='utf8') as f:
            raw = f.read()
    except FileNotFoundError:
        return {'file': path}

    config_reader = configparser.ConfigParser()
    config_reader.read_string(raw)

    for section in ['migrations', 'templates', 'database']:
        if section not in config_reader:
            print("Config has no {!r} section".format(section), file=sys.stderr)
            sys.exit(1)

    config = {
        'file': path,
        'migrations': dict(config_reader.items('migrations')),
        'templates': dict(config_reader.items('templates')),
        'database': dict(config_reader.items('database')),
    }

    config['migrations']['path'] = migranite.utils.parse_path(config['migrations']['path'])
    config['migrations']['digits'] = int(config['migrations'].get('digits', 3))
    config['templates']['path'] = migranite.utils.parse_path(config['templates']['path'])

    return config


def main():
    parser = argparse.ArgumentParser(add_help=False)

    parser.add_argument('--config',
                        type=_parse_config,
                        default='.migranite',
                        help="config file (default '.migranite')")

    subparsers = parser.add_subparsers()

    # init

    parser_init = subparsers.add_parser('init', help="add migranite to project")
    parser_init.set_defaults(func=run_init)

    parser_init.add_argument('--migrations',
                             type=migranite.utils.parse_path,
                             required=True,
                             metavar='DIR',
                             help="migration directory")

    parser_init.add_argument('--templates',
                             type=migranite.utils.parse_path,
                             required=True,
                             metavar='DIR',
                             help="templates directory")

    # list

    parser_list = subparsers.add_parser('list', help="show all available migrations")
    parser_list.set_defaults(func=run_list)

    parser_list.add_argument('-l', '--long',
                             action='store_true',
                             help="show long descriprions")

    parser_list.add_argument('-a', '--all',
                             action='store_true',
                             help="show removed migrations migrations too")

    # create

    parser_create = subparsers.add_parser('create', help="create new migtation")
    parser_create.set_defaults(func=run_create)

    parser_create.add_argument('name',
                               nargs=1,
                               metavar='NAME',
                               help="new migration name")

    parser_create.add_argument('-t', '--template',
                               default=None,
                               metavar='NAME',
                               help="template for new migration")

    # run

    parser_run = subparsers.add_parser('run', help="run migrations")
    parser_run.set_defaults(func=run)

    parser_run.add_argument('migrations',
                            nargs='*',
                            help="migrations for run")

    parser_run.add_argument('-f', '--force',
                            action='store_true',
                            help="force run specified migrations")

    # help

    parser_help = subparsers.add_parser('help', help="show this help message and exit")
    parser_help.set_defaults(func=run_help)

    # version

    parser_version = subparsers.add_parser('version', help="show version and exit")
    parser_version.set_defaults(func=run_version)

    # --

    args = parser.parse_args()
    getattr(args, 'func', run_help)(parser, args)


if __name__ == '__main__':
    main()



import abc
import os
import hashlib


class MigrationInterface(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def __init__(self, config, file_name, verbose=False):
        pass

    @abc.abstractproperty
    def config(self):
        """ Config dict
        """

    @abc.abstractproperty
    def file_name(self):
        """ Migration file name
        """

    @abc.abstractproperty
    def verbose(self):
        """ Verbose flag
        """

    @abc.abstractproperty
    def num(self):
        """ Number representation of migration
        """

    @abc.abstractproperty
    def name(self):
        """ Name of migration
        """

    @abc.abstractproperty
    def source(self):
        """ Migration source code
        """

    @abc.abstractproperty
    def md5(self):
        """ Migration source code md5
        """

    @abc.abstractproperty
    def short(self):
        """ One-line description
        """

    @abc.abstractproperty
    def long(self):
        """ Multiline long description
        """

    @abc.abstractmethod
    def run(self):
        """ Migration logic
        """

    @abc.abstractmethod
    def __lt__(self, other):
        pass

    @abc.abstractmethod
    def __eq__(self, other):
        pass


class MigrationBase(MigrationInterface):
    _source = None

    def __init__(self, config, file_name, verbose=False):
        self._config = config
        self._file_name = file_name
        self._verbose = verbose

        file_base_name = os.path.splitext(file_name)[0]
        num, name = file_base_name.split('-', 1)
        self._num = int(num)
        self._name = name

    @property
    def config(self):
        return self._config

    @property
    def file_name(self):
        return self._file_name

    @property
    def verbose(self):
        return self._verbose

    @property
    def num(self):
        return self._num

    @property
    def name(self):
        return self._name

    @property
    def source(self):
        if self._source is None:
            f = open(os.path.join(self.config['migrations']['path'], self.file_name), 'r')
            with f:
                self._source = f.read()
        return self._source

    @property
    def md5(self):
        return hashlib.md5(self.source.encode('utf8')).hexdigest()

    def __lt__(self, other):
        if self.num == other.num:
            return self.name < other.name
        else:
            return self.num < other.num

    def __eq__(self, other):
        return self.num == other.num and self.name == other.name

    def __repr__(self):
        return "<{} {!s}>".format(self.__class__.__name__, self.file_name)

    def __str__(self):
        return "{m.num}-{m.name}: {m.short}".format(m=self)



import sys

from datetime import datetime

try:
    import pymongo
    PYMONGO = True
except ImportError:
    PYMONGO = False

from .base import DBInterface, MigrationResult


class Mongo(DBInterface):
    _collection = None

    def __init__(self, config):
        if not PYMONGO:
            print("pymongo not installed.", file=sys.stderr)
            sys.exit(1)

        if not config['database'].get('name'):
            print("Database name not set in config.", file=sys.stderr)
            sys.exit(1)

        self._config = config
        self.host = config['database'].get('host', 'localhost')
        self.port = int(config['database'].get('port', 27017))
        self.db_name = config['database']['name']
        self.collection_name = config['database'].get('collection', 'migrations')

    @property
    def collection(self):
        if self._collection is None:
            client = pymongo.MongoClient(self.host, self.port)
            db = client[self.db_name]
            self._collection = db[self.collection_name]
            self._collection.create_index([('ts', 1)])
            self._collection.create_index([('num', 1), ('name', 1)])

        return self._collection

    @property
    def results(self):
        docs = self.collection.find().sort([('ts', 1)])
        return [self._result_from_doc(doc) for doc in docs]

    def add(self, migration, result):
        doc = self._doc_from_migration(migration, result)
        self.collection.insert(doc)

    def find(self, migration):
        docs = self.collection.find({
            'num': migration.num,
            'name': migration.name,
        }).sort([('ts', 1)])
        return [self._result_from_doc(doc) for doc in docs]

    def success(self, migration):
        results = self.find(migration)
        return results and results[-1].result

    @staticmethod
    def _doc_from_migration(migration, result):
        return {
            'ts': datetime.now(),
            'num': migration.num,
            'name': migration.name,
            'short': migration.short,
            'long': migration.long,
            'source': migration.source,
            'md5': migration.md5,
            'result': result,
        }

    @staticmethod
    def _result_from_doc(doc):
        return MigrationResult(
            ts=doc['ts'],
            num=doc['num'],
            name=doc['name'],
            short=doc['short'],
            long=doc['long'],
            source=doc['source'],
            md5=doc['md5'],
            result=doc['result'],
        )



from string import Template

CONFIG = Template("""\
[migrations]
path = $migranite_path
; digits = 3

[templates]
path = $templates_path
default = default.py

[database]
backend = json
path = .migranite_db.json

; [database]
; backend = mongo
; host = localhost
; port = 27017
; name =
; collection = migrations
""")

TEMPLATE_DEFAULT = '''\
""" <EMPTY DESCRIPTION>

<EMPTY LONG DESCRIPTION>
"""


def run():
    pass
'''



import os

from .base import MigrationInterface  # noqa
from .py import MigrationPy


def get_all(config):
    """ Get all available migtarion
    """
    return [get(config, fn) for fn in get_files(config)]


def get_files(config):
    path = config['migrations']['path']
    if not os.path.isdir(path):
        print("{} not exists".format(path))
        return []

    for file_name in sorted(os.listdir(path)):
        if file_name.startswith('.'):
            continue
        elif not os.path.splitext(file_name)[1]:
            continue

        yield file_name


def get(config, file_name):
    """ Get migration object for file
    """
    name, ext = os.path.splitext(file_name)

    if ext == '.py':
        return MigrationPy(config, file_name)
    else:
        raise ValueError("unknown migration type {!r}".format(file_name))


def find(migrations, num=None, name=None):
    """ Find migtations for num and name
    """
    result = []

    for migration in migrations:
        if num is not None and name is not None:
            if migration.num == num and migration.name == name:
                result.append(migration)
        elif num is not None:
            if migration.num == num:
                result.append(migration)
        elif name is not None:
            if migration.name == name:
                result.append(migration)
        else:
            result.extend(migrations)

    return result



import os
import sys

from colorama import Fore, Style

from . import migrations
from .db import get as get_db


def init(config, migrations, templates):
    from . import init_data

    config_path = config['file']
    config_dir = os.path.dirname(config_path)
    os.makedirs(config_dir, exist_ok=True)

    with open(config_path, 'w') as f:
        data = init_data.CONFIG.safe_substitute(
            migranite_path=migrations,
            templates_path=templates,
        )
        f.write(data)

    os.makedirs(migrations, exist_ok=True)
    os.makedirs(templates, exist_ok=True)

    templates_default_path = os.path.join(templates, 'default.py')
    if not os.path.exists(templates_default_path):
        with open(templates_default_path, 'w') as f:
            f.write(init_data.TEMPLATE_DEFAULT)
    else:
        print("{!r} is already exist. Skip.".format(templates_default_path),
              file=sys.stderr)


def print_list(config, long=False, all=False):
    """ Print list of available migrations

    :config: config dict
    """
    db = get_db(config)

    def print_item(item, status, long=False):
        if status == 'PSS':
            color = Fore.GREEN
        elif status == 'ERR':
            color = Fore.RED
        else:
            color = ''

        print("[{C}{status}{R}] {item.num}-{item.name}: {item.short}".format(
            status=status, item=item, C=color, R=Style.RESET_ALL))

        if long and item.long:
            print("\n{}\n".format(item.long))

    if all:
        for result in db.results:
            print_item(result, 'PSS' if result.result else 'ERR', long=long)

    for item in migrations.get_files(config):
        item = migrations.get(config, item)

        if db.success(item):
            status = 'PSS'
        elif db.find(item):
            status = 'ERR'
        else:
            status = '   '

        if all and status.strip():
            continue

        print_item(item, status, long=long)


def create(config, template, name):
    """ Create new migration

    :temlate: template name
    :name: migration name
    """
    with open(template) as f:
        data = f.read()

    digits = config.get('digits', 3)
    _, ext = os.path.splitext(template)

    num = max([m.num for m in migrations.get_all(config)] + [0]) + 1

    name = '{}-{}{}'.format(str(num).zfill(digits), name, ext)

    path = os.path.join(config['migrations']['path'], name)

    with open(path, 'w') as f:
        f.write(data)


def migrate(config, migrations_names, force=False):
    """ Run specified migrations

    :config: config dict
    :migrations_names: list of migrations names
    :force: run if migration already worked (default False)
    """
    db = get_db(config)
    objects = migrations.get_all(config)
    found = []

    for mn in migrations_names:
        if '-' in mn and mn.split('-', 1)[0].isdigit():
            num, name = mn.split('-', 1)
            if num.isdigit():
                num = int(num)

        elif mn.isdigit():
            num, name = int(mn), None

        else:
            num, name = None, mn

        found_migrations = migrations.find(objects, num, name)

        if not found_migrations:
            print("Migration {!r} not found".format(mn),
                  file=sys.stderr)
            sys.exit(1)
        elif len(found_migrations) > 1:
            print("More than one migration found for {!r}".format(mn),
                  file=sys.stderr)
            sys.exit(1)
        else:
            migration = found_migrations[0]
            if force or not db.success(migration):
                found.append(migration)
            else:
                print("[{S}SKP{R}] {!s}".format(
                    migration, S=Fore.YELLOW, R=Style.RESET_ALL))

    for migration in sorted(found):
        _run_migration(db, migration)


def migrate_all(config):
    """ Run all migrations

    :config: config dict
    """
    db = get_db(config)
    for migration in migrations.get_all(config):
        if not db.success(migration):
            _run_migration(db, migration)
        else:
            print("[{S}SKP{R}] {!s}".format(
                migration, S=Fore.YELLOW, R=Style.RESET_ALL))


def _run_migration(db, migration):
    print("[{S}RUN{R}] {!s}".format(
        migration, S=Fore.GREEN, R=Style.RESET_ALL))

    try:
        result = False
        migration.run()
        result = True
    finally:
        db.add(migration, result)
        if not result:
            print("[{S}ERR{R}] {!s}".format(
                migration, S=Fore.RED, R=Style.RESET_ALL))



from .json import JSON
from .mongo import Mongo


def get(config):
    """ Get database object
    """
    if config['database']['backend'] == 'json':
        return JSON(config)
    elif config['database']['backend'] == 'mongo':
        return Mongo(config)
    else:
        raise RuntimeError("Bad database backend: {}".format(config['database']))



import sys

import json
from datetime import datetime

import dateutil.parser

from .base import DBSimpleBase, MigrationResult


class JSON(DBSimpleBase):
    _results = None

    def __init__(self, config):
        super().__init__(config)
        if 'path' not in config['database']:
            print("Path to json file not set in settings.", file=sys.stderr)
            sys.exit(1)

        self.path = config['database']['path']

    @property
    def results(self):
        if self._results is None:
            try:
                with open(self.path, 'r', encoding='utf8') as f:
                    raw = json.load(f)
            except FileNotFoundError:
                raw = []
            except json.decoder.JSONDecodeError:
                raise RuntimeError("Bad {} format".format(self.path))

            results = []

            for raw_obj in raw:
                result = self._result_from_json(raw_obj)
                results.append(result)

            self._results = results

        return self._results

    def add(self, migration, result):
        r = self._result_from_migration(migration, result)
        self.results.append(r)
        self._save_results()

    @staticmethod
    def _result_from_migration(migration, result):
        return MigrationResult(
            ts=datetime.now(),
            num=migration.num,
            name=migration.name,
            short=migration.short,
            long=migration.long,
            source=migration.source,
            md5=migration.md5,
            result=result,
        )

    @staticmethod
    def _result_from_json(raw):
        return MigrationResult(
            ts=dateutil.parser.parse(raw['ts']),
            num=int(raw['num']),
            name=raw['name'],
            short=raw['short'],
            long=raw['long'],
            source=raw['source'],
            md5=raw['md5'],
            result=raw['result'],
        )

    @staticmethod
    def _result_to_json(result):
        return {
            'ts': result.ts.isoformat(sep='T'),
            'num': result.num,
            'name': result.name,
            'short': result.short,
            'long': result.long,
            'source': result.source,
            'md5': result.md5,
            'result': result.result,
        }

    def _save_results(self):
        raw = [self._result_to_json(r) for r in self.results]
        with open(self.path, 'w', encoding='utf8') as f:
            json.dump(raw, f, indent=2)



import types

from .base import MigrationBase


class MigrationPy(MigrationBase):
    _source = None
    _module = None
    _short = None
    _long = None

    @property
    def short(self):
        if self._short is None:
            self._read_doc()
        return self._short

    @property
    def long(self):
        if self._long is None:
            self._read_doc()
        return self._long

    def run(self):
        self.module.run()

    @property
    def module(self):
        if self._module is None:
            self._read_module()
        return self._module

    def _read_module(self):
        self._module = types.ModuleType(self.file_name)
        exec(self.source, self.module.__dict__)

    def _read_doc(self):
        raw = self.module.__doc__.strip()
        if '\n' in raw:
            t = tuple(s.strip() for s in raw.split('\n', 1))
            self._short, self._long = t
        else:
            self._short, self._long = raw, ''



from setuptools import setup, find_packages

setup(
    name='migranite',
    description='Migranite -- manage migtations tool',
    version='0.4.0',
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Topic :: Software Development",
    ],
    url='https://github.com/zzzsochi/migranite',
    keywords=['migrations', 'development'],
    packages=find_packages(),
    install_requires=['python-dateutil', 'colorama'],
    extras_require={'mongo': ['pymongo']},
    entry_points={
        'console_scripts': [
            'migranite = migranite.__main__:main',
        ],
    },
)



import os


def parse_path(path):
    if ':' in path:
        package_name, name = path.split(':', 1)
        if not name:
            raise ValueError("Empty config file name")

        try:
            package = __import__(package_name, fromlist=package_name.split('.'))
        except ImportError as exc:
            raise ValueError from exc

        package_dir = os.path.dirname(package.__file__)
        return os.path.join(package_dir, name)
    else:
        return os.path.abspath(os.path.expanduser(path))
